"""
Heckman correction for sample selection bias (the Heckman procedure).

This is a modified version of the code available at
https://github.com/statsmodels/statsmodels/blob/92ea62232fd63c7b60c60bee4517ab3711d906e3/statsmodels/regression/Heckit.py

Created by Tetsu Haruyama
Last Modified 2 July 2019
"""

import numpy as np
import statsmodels.api as sm
import statsmodels.base.model as base
from statsmodels.iolib import summary
from statsmodels.tools.numdiff import approx_fprime
from scipy.stats import norm
import warnings
import pandas as pd
from statsmodels.tools.decorators import cache_readonly, cache_writable    # Tetsu
from statsmodels.compat.python import lrange

class Heckit(base.LikelihoodModel):
    """
    Class for Heckit correction for sample selection bias model.

    Parameters
    ----------
    endog : 1darray
        Data for the dependent variable. Should be set to np.nan for
        censored observations.
    exog : 2darray
        Data for the regression (response) equation. If a constant
        term is desired, the user should directly add the constant
        column to the data before using it as an argument here.
    exog_select : 2darray
        Data for the selection equation. If a constant
        term is desired, the user should directly add the constant
        column to the data before using it as an argument here.
    **kwargs:
        missing=, which can be 'none', 'drop', or 'raise'

    Notes
    -----
    The selection equation should contain at least one variable that
    is not in the regression (response) equation, i.e. the selection
    equation should contain at least one instrument. However, this
    module should still work if the user chooses not to do this.
    """

    def __init__(self, endog, exog, exog_select, **kwargs):

        # check that Z has same index as X (and consequently Y through super().__init__)
        if pd.__name__ in type(endog).__module__ and pd.__name__ in type(exog).__module__:
            if not all(endog.index==exog_select.index):
                raise ValueError("Z indices need to be the same as X and Y indices")


        # shape checks
        if (len(endog) == len(exog)) and (len(endog) == len(exog_select)):
            pass
        else:
            raise ValueError("Y, X, and Z data shapes do not conform with each other.")

        try:
            if (endog.ndim == 1) and (exog.ndim <= 2) and (exog_select.ndim <= 2):
                pass
            else:
                raise ValueError("Y, X, and Z data shapes do not conform with each other.")
        except AttributeError:
            if (np.asarray(endog).ndim == 1) and (np.asarray(exog).ndim <= 2) and (np.asarray(exog_select).ndim <= 2):
                pass
            else:
                raise ValueError("Y, X, and Z data shapes do not conform with each other.")

        # give missing (treated) values in endog variable finite values so that super().__init__
        # does not strip them out -- they will be put back after the call to super().__init__
        treated = np.asarray(~np.isnan(endog))

        try:
            endog_nomissing = endog.copy()
            endog_nomissing[~treated] = -99999
        except (TypeError, AttributeError):
            endog_nomissing = [endog[i] if treated[i] else -99999 for i in range(len(treated))]

        # create 1-D array that will be np.nan for every row of exog_select that has any missing
        # values and a finite value otherwise for the call to super().__init__ so that it can
        # strip out rows where exog_select has missing data if missing option is set

        if np.asarray(exog_select).ndim==2:
            exog_select_1dnan_placeholder = \
                [np.nan if any(np.isnan(row)) else 1 for row in np.asarray(exog_select)]
        else:  # assume ==1
            exog_select_1dnan_placeholder = [np.nan if np.isnan(row) else 1 for row in np.asarray(exog_select)]

        if pd.__name__ in type(endog).__module__:
            exog_select_1dnan_placeholder = pd.Series(exog_select_1dnan_placeholder, index=endog.index)
        else:
            exog_select_1dnan_placeholder = np.array(exog_select_1dnan_placeholder)

        # create array of sequential row positions so that rows of exog_select that have missing
        # data can be identified after call to super().__init__
        obsno = np.array(list(range(len(endog))))

        # call super().__init__
        super(Heckit, self).__init__(
            endog_nomissing, exog=exog,
            exog_select_1dnan_placeholder=exog_select_1dnan_placeholder, obsno=obsno,
            treated=treated,
            **kwargs)

        # put np.nan back into endog for treated rows
        self.endog = self.data.endog = \
            np.asarray(
                [self.endog[i] if self.treated[i] else np.nan for i in range(len(self.treated))]
                )

        # strip out rows stripped out by call to super().__init__ in Z variable
        self.exog_select = np.asarray([np.asarray(exog_select)[obs] for obs in self.obsno])

        # store variable names of exog_select
        try:
            self.exog_select_names = exog_select.columns.tolist()
        except AttributeError:
            self.exog_select_names = None

        # delete attributes created by the call to super().__init__ that are no longer needed
        del self.exog_select_1dnan_placeholder
        del self.obsno


        # store observation counts
        self.nobs_total = len(endog)
        self.nobs_uncensored = self.nobs = np.sum(self.treated)
        self.nobs_censored = self.nobs_total - self.nobs_uncensored


    def initialize(self):
        self.wendog = self.endog
        self.wexog = self.exog
        self._df_model = None   # Tetsu
        self.rank = None    # Tetsu
        self._df_resid = None # Tetsu


    def whiten(self, data):
        """
        Model whitener for Heckit correction model does nothing.
        """
        return data


    def get_datamats(self):
        Y = np.asarray(self.endog)
        Y = Y[self.treated]

        X = np.asarray(self.exog)
        X = X[self.treated,:]
        if X.ndim==1:
            X = np.atleast_2d(X).T

        Z = np.asarray(self.exog_select)
        if Z.ndim==1:
            Z = np.atleast_2d(Z).T

        return Y, X, Z


    # Tetsu ------ starts -----
    @property
    def df_model(self):
        """
        The model degree of freedom, defined as the rank of the regressor
        matrix minus 1 if a constant is included.
        """
        X = np.asarray(self.exog)
        X = X[self.treated,:]
        if self._df_model is None:
            if self.rank is None:
                self.rank = np.linalg.matrix_rank(X)
                self._df_model = float(self.rank - self.k_constant)
        return self._df_model


    @df_model.setter
    def df_model(self, value):
        self._df_model = value

    @property
    def df_resid(self):
        """
        The residual degree of freedom, defined as the number of observations
        minus the rank of the regressor matrix.
        """
        X = np.asarray(self.exog)
        X = X[self.treated,:]

        if self._df_resid is None:
            if self.rank is None:
                self.rank = np.linalg.matrix_rank(X)
            self._df_resid = self.nobs - self.rank
        return self._df_resid

    @df_resid.setter
    def df_resid(self, value):
        self._df_resid = value

    # Tetsu ----- ends -----





    def fit(self, method='twostep',
            cov_type_1='nonrobust', cov_type_2='nonrobust',  #  Tetsu ################
             start_params_mle=None, method_mle=None, maxiter_mle=None, **kwargs_mle):
        """
        Fit the Heckit selection model.

        Parameters
        ----------
        method : str
            Can only be "2step", which uses Heckit's two-step method.
        start_params_mle: 1darray
            If using MLE, starting parameters.
        method_mle: str
            If using MLE, the MLE estimation method.
        maxiter_mle: scalar
            If using MLE, the maximum number of iterations for MLE estimation.
        **kwargs_mle:
            Other arguments to pass to optimizer for MLE estimation.

        Returns
        -------
        A HeckitResults class instance.

        See Also
        ---------
        HeckitResults


        """

        ## Show warning to user if estimation is by two-step but MLE arguments were also provided
        if method=='twostep':
            if start_params_mle is not None or method_mle is not None or maxiter_mle is not None or \
                len(kwargs_mle.keys())>0:
                warnings.warn('The user chose to estimate the Heckit model by Two-Step,' + \
                    ' but MLE arguments were provided. Extraneous MLE arguments will be ignored.')

        results = self._fit_twostep(cov_type_1=cov_type_1,cov_type_2=cov_type_2)  # Tetsu : cov_type_1=cov_type_1,cov_type_2=cov_type_2
        return results


    def _fit_twostep(self,cov_type_1='nonrobust',cov_type_2='nonrobust'):  # Tetsu : cov_type_1=cov_type_1,cov_type_2=cov_type_2
        ########################################################################
        # PRIVATE METHOD
        # Fits using Heckit two-step from Heckit (1979).
        ########################################################################

        ## prep data
        Y, X, Z = self.get_datamats()

        ## Step 1
        step1model = sm.Probit(self.treated, Z)
        step1res = step1model.fit(disp=False,cov_type=cov_type_1)  # Tets cov_type=cov_type_1
        step1_fitted = np.atleast_2d(step1res.fittedvalues).T
        step1_varcov = step1res.cov_params()

        inverse_mills = norm.pdf(step1_fitted)/norm.cdf(step1_fitted)

        ## Step 2
        W = np.hstack((X, inverse_mills[self.treated] ) )
        step2model = sm.OLS(Y, W)
        step2res = step2model.fit(cov_type=cov_type_2)    # Tetsu : cov_type_2=cov_type

        params = step2res.params[:-1]
        betaHat_inverse_mills = step2res.params[-1]


        ## Compute standard errors
        # Compute estimated error variance of censored regression
        delta = np.multiply(inverse_mills, inverse_mills + step1_fitted)[self.treated]

        sigma2Hat = step2res.resid.dot(step2res.resid) / self.nobs_uncensored + \
            (betaHat_inverse_mills**2 * sum(delta)) / self.nobs_uncensored
        sigma2Hat = sigma2Hat[0]
        sigmaHat = np.sqrt(sigma2Hat)
        rhoHat = betaHat_inverse_mills / sigmaHat

        # compute standard errors of beta estimates of censored regression
        delta_1d = delta.T[0]

        Q = rhoHat**2 * ((W.T*delta_1d).dot(Z[self.treated])).dot(step1_varcov).dot((Z[self.treated].T*delta_1d).dot(W))

        WT_W_inv = np.linalg.inv(W.T.dot(W))
        WT_R = W.T*(1 - rhoHat**2 * delta_1d)
        normalized_varcov_all = WT_W_inv.dot(WT_R.dot(W)+Q).dot(WT_W_inv)
        del WT_W_inv
        del WT_R

        del delta_1d

        normalized_varcov = normalized_varcov_all[:-1,:-1]

        varcov_all = sigma2Hat * normalized_varcov_all
        varcov = varcov_all[:-1,:-1]

        stderr_all = np.sqrt(np.diag(varcov_all))
        stderr = stderr_all[:-1]
        stderr_betaHat_inverse_mills = stderr_all[-1]


        ## store results
        results = HeckitResults(self, params, normalized_varcov, sigma2Hat,
            select_res=step1res,
            params_inverse_mills=betaHat_inverse_mills, stderr_inverse_mills=stderr_betaHat_inverse_mills,
            var_reg_error=sigma2Hat, corr_eqnerrors=rhoHat,
            method='twostep', cov_type_1=cov_type_1, cov_type_2=cov_type_2)  #  cov_type_1=cov_type_1, cov_type_2=cov_type_2

        return results


    def predict(self, params, exog=None):
        """
        Return linear predicted values from a design matrix.

        Parameters
        ----------
        exog : array-like
            Design / exogenous data
        params : array-like, optional after fit has been called
            Parameters of a linear model

        Returns
        -------
        An array of fitted values

        Notes
        -----
        If the model has not yet been fit, params is not optional.
        """
        if exog is None:
            exog = self.exog
        return np.dot(exog, params)

        if self._results is None and params is None:
            raise ValueError("If the model has not been fit, then you must specify the params argument.")
        if self._results is not None:
            return np.dot(exog, self._results.params)
        else:
            return np.dot(exog, params)


class HeckitResults(base.LikelihoodModelResults):
    """
    Class to represent results/fits for Heckit model.

    Attributes
    ----------
    select_res : ProbitResult object
        The ProbitResult object created when estimating the selection equation.
    params_inverse_mills : scalar
        Parameter estimate of the coef on the inverse Mills term in the second step.
    stderr_inverse_mills : scalar
        Standard error of the parameter estimate of the coef on the inverse Mills
        term in the second step.
    var_reg_error : scalar
        Estimate of the "sigma" term, i.e. the error variance estimate of the
        regression (response) equation
    corr_eqnerrors : scalar
        Estimate of the "rho" term, i.e. the correlation estimate of the errors between the
        regression (response) equation and the selection equation
    method : string
        The method used to produce the estimates, i.e. 'twostep', 'mle'
    """

    def __init__(self, model, params, normalized_cov_params=None, scale=1.,
        select_res=None,
        params_inverse_mills=None, stderr_inverse_mills=None,
        var_reg_error=None, corr_eqnerrors=None,
        method=None, cov_type_1=None, cov_type_2=None):  # Tetsu : cov_type_1=None, cov_type_2=None

        super(HeckitResults, self).__init__(model, params,
                                                normalized_cov_params,
                                                scale)

        self.select_res = select_res
        self.params_inverse_mills = params_inverse_mills
        self.stderr_inverse_mills = stderr_inverse_mills
        self.var_reg_error = var_reg_error
        self.corr_eqnerrors = corr_eqnerrors
        self.method = method


    # Tetsu ----- starts ------
        self.cov_type_1 = cov_type_1
        self.cov_type_2 = cov_type_2
        self.df_model = model.df_model
        self.df_resid = model.df_resid
        dropnan = np.asarray(~np.isnan(self.model.endog))
        self.endog = self.model.endog[dropnan]
        self.exog = self.model.exog[dropnan]
        self.wendog = self.endog
        self.wexog = self.exog
        self.pinv_wexog = np.linalg.pinv(self.exog)

        if not hasattr(self, 'use_t'):
            self.use_t = False
            # self.use_t = True

        if not hasattr(self.select_res, 'use_t'):
            self.select_res.use_t = False

    @cache_readonly
    def nobs(self):
        """Number of observations n."""
        # return float(self.model.nobs_total)
        return float(len(self.endog))

    @cache_readonly
    def fittedvalues(self):
        """The predicted values for the original (unwhitened) design."""
        # return self.model.predict(self.params, self.model.exog)
        return self.model.predict(self.params, self.exog)

    @cache_readonly
    def resid(self):
        """The residuals of the model."""
        return self.endog - self.model.predict(self.params, self.exog)

    @cache_readonly
    def wresid(self):
        """
        The residuals of the transformed/whitened regressand and regressor(s)
        """
        return self.wendog - self.model.predict(self.params, self.wexog)

    @cache_readonly
    def ssr(self):
        """Sum of squared (whitened) residuals."""
        wresid = self.wresid
        return np.dot(wresid, wresid)

    @cache_readonly
    def centered_tss(self):
        """The total (weighted) sum of squares centered about the mean."""
        model = self.model
        weights = getattr(model, 'weights', None)
        sigma = getattr(model, 'sigma', None)
        if weights is not None:
            # mean = np.average(model.endog, weights=weights)
            mean = np.average(self.endog, weights=weights)
            return np.sum(weights * (self.endog - mean)**2)

        elif sigma is not None:
            # Exactly matches WLS when sigma is diagonal
            iota = np.ones_like(self.endog)
            iota = model.whiten(iota)
            mean = self.wendog.dot(iota) / iota.dot(iota)
            err = self.endog - mean
            err = model.whiten(err)
            return np.sum(err**2)
        else:
            centered_endog = self.wendog - self.wendog.mean()
            return np.dot(centered_endog, centered_endog)

    @cache_readonly
    def uncentered_tss(self):
        """
        Uncentered sum of squares.  Sum of the squared values of the
        (whitened) endogenous response variable.
        """
        wendog = self.wendog
        return np.dot(wendog, wendog)

    @cache_readonly
    def ess(self):
        """Explained sum of squares. If a constant is present, the centered
        total sum of squares minus the sum of squared residuals. If there is no
        constant, the uncentered total sum of squares is used."""

        if self.k_constant:
            return self.centered_tss - self.ssr
        else:
            return self.uncentered_tss - self.ssr

    @cache_readonly
    def rsquared(self):
        """
        R-squared of a model with an intercept.  This is defined here
        as 1 - `ssr`/`centered_tss` if the constant is included in the
        model and 1 - `ssr`/`uncentered_tss` if the constant is
        omitted.
        """
        if self.k_constant:
            return 1 - self.ssr/self.centered_tss
        else:
            return 1 - self.ssr/self.uncentered_tss

    @cache_readonly
    def rsquared_adj(self):
        """
        Adjusted R-squared.  This is defined here as 1 -
        (`nobs`-1)/`df_resid` * (1-`rsquared`) if a constant is
        included and 1 - `nobs`/`df_resid` * (1-`rsquared`) if no
        constant is included.
        """
        return 1 - (np.divide(self.nobs - self.k_constant, self.df_resid)
                    * (1 - self.rsquared))

    @cache_readonly
    def mse_model(self):
        """
        Mean squared error the model. This is the explained sum of
        squares divided by the model degrees of freedom.
        """
        return self.ess/self.df_model


    @cache_readonly
    def mse_resid(self):
        """
        Mean squared error of the residuals.  The sum of squared
        residuals divided by the residual degrees of freedom.
        """
        return self.ssr/self.df_resid


    @cache_readonly
    def mse_total(self):
        """
        Total mean squared error.  Defined as the uncentered total sum
        of squares divided by n the number of observations.
        """
        if self.k_constant:
            return self.centered_tss / (self.df_resid + self.df_model)
        else:
            return self.uncentered_tss / (self.df_resid + self.df_model)


    @cache_readonly
    def fvalue(self):
        """F-statistic of the fully specified model.  Calculated as the mean
        squared error of the model divided by the mean squared error of the
        residuals."""
        if hasattr(self, 'cov_type_1') and self.cov_type_1 != 'nonrobust':
            # with heteroscedasticity or correlation robustness
            k_params = self.normalized_cov_params.shape[0]
            mat = np.eye(k_params)
            const_idx = self.model.data.const_idx
            # TODO: What if model includes implicit constant, e.g. all
            #       dummies but no constant regressor?
            # TODO: Restats as LM test by projecting orthogonalizing
            #       to constant?
            if self.model.data.k_constant == 1:
                # if constant is implicit, return nan see #2444
                if const_idx is None:
                    return np.nan

                idx = lrange(k_params)
                idx.pop(const_idx)
                mat = mat[idx]  # remove constant
                if mat.size == 0:  # see  #3642
                    return np.nan
            ft = self.f_test(mat)
            # using backdoor to set another attribute that we already have
            self._cache['f_pvalue'] = ft.pvalue
            return ft.fvalue
        else:
            # for standard homoscedastic case
            return self.mse_model/self.mse_resid


    @cache_readonly
    def f_pvalue(self):
        """p-value of the F-statistic"""
        from scipy import stats
        return stats.f.sf(self.fvalue, self.df_model, self.df_resid)


    def _HCCM(self, scale):
        H = np.dot(self.pinv_wexog,
                   scale[:, None] * self.pinv_wexog.T)
        return H

    @cache_readonly
    def cov_HC1(self):
        """
        See statsmodels.RegressionResults
        """

        self.het_scale = self.nobs/(self.df_resid)*(self.wresid**2)
        cov_HC1 = self._HCCM(self.het_scale)
        return cov_HC1

    @cache_readonly
    def HC1_se(self):
        """
        See statsmodels.RegressionResults
        """
        return np.sqrt(np.diag(self.cov_HC1))

    # Tetsu ----- ends -----



    def summary(self, yname=None, xname=None, zname=None, title=None, alpha=.05):
        """Summarize the Heckit model Results

        Parameters
        -----------
        yname : string, optional
            Default is `y`
        xname : list of strings, optional
            Default is `x_##` for ## in p the number of regressors
            in the regression (response) equation.
        zname : list of strings, optional
            Default is `z_##` for ## in p the number of regressors
            in the selection equation.
        title : string, optional
            Title for the top table. If not None, then this replaces the
            default title
        alpha : float
            significance level for the confidence intervals

        Returns
        -------
        smry : Summary instance
            this holds the summary tables and text, which can be printed or
            converted to various output formats.

        See Also
        --------
        statsmodels.iolib.summary.Summary : class to hold summary
            results

        """

        ## Put in Z name detected from data if none supplied, unless that too could not be
        ## inferred from data, then put in generic names
        if zname is None and self.model.exog_select_names is not None:
            zname=self.model.exog_select_names
        elif zname is None and self.model.exog_select_names is None:
            try:
                zname = ['z' + str(i) for i in range(len(self.model.exog_select[0]))]
                zname[0]  = 'z0_or_zconst'
            except TypeError:
                zname = 'z0_or_zconst'

        try:  # for Python 3
            if isinstance(zname, str):
                zname = [zname]
        except NameError:  # for Python 2
            if isinstance(zname, basestring):
                zname = [zname]


        ## create summary object
        # instantiate the object
        smry = summary.Summary()

        top_left = [('Dep. Variable:', None),
                    ('Model:', None),
                    ('Method:', ['Heckman Two-Step']),
                    ('Date:', None),
                    ('Time:', None),
                    ('No. Total Obs.:', ["%#i" % self.model.nobs_total]),
                    ('No. Censored Obs.:', ["%#i" % self.model.nobs_censored]),
                    ('No. Uncensored Obs.:', ["%#i" % self.model.nobs_uncensored]),
                    ]


        top_right = [('R-squared:', ["%#8.3f" % self.rsquared]),
                     ('Adj. R-squared:', ["%#8.3f" % self.rsquared_adj]),
                     ('F-statistics:', ["%#8.3f" % self.fvalue]),
                     ('Prob (F-statistic):', ["%#8.3f" % self.f_pvalue]),  # Tetsu added
                     ('Cov in 1st Stage:', [self.cov_type_1]),  # Tetsu added
                     ('Cov in 2nd Stage:', [self.cov_type_2]),  # Tetsu added
                     ]

        if title is None:
            title = self.model.__class__.__name__ + ' ' + "Regression Results"

        smry.add_table_2cols(self, gleft=top_left, gright=top_right,
                          yname=yname, xname=xname, title=title)

        # add the Heckit-corrected regression table
        smry.add_table_params(self, yname=yname, xname=xname, alpha=alpha,
                             use_t=self.use_t)

        # add the selection equation estimates table
        smry.add_table_params(self.select_res, yname=yname, xname=zname, alpha=alpha,
                             use_t=self.select_res.use_t)

        # add the estimate to the inverse Mills estimate (z-score)
        smry.add_table_params(
            base.LikelihoodModelResults(None, np.atleast_1d(self.params_inverse_mills),
            normalized_cov_params=np.atleast_1d(self.stderr_inverse_mills**2), scale=1.),
            yname=None, xname=['IMR (Lambda)'], alpha=alpha,
            use_t=False)

        # add point estimates for rho and sigma
        diagn_left = [('rho:', ["%#6.3f" % self.corr_eqnerrors]),
                      ('sigma:', ["%#6.3f" % np.sqrt(self.var_reg_error)]),
                      ]

        diagn_right = [
                       ]

        smry.add_table_2cols(self, gleft=diagn_left, gright=diagn_right,
                          yname=yname, xname=xname,
                          title="")

        # add text at end
        smry.add_extra_txt(['First table are the estimates for the regression (response) equation.',
            'Second table are the estimates for the selection equation.',
            'Third table is the estimate for the coef of the inverse Mills ratio (Heckman\'s Lambda).'])

        ## return
        return smry



"""
Created by Tetsu Haruyama
Last modified 15 July 2019
"""

import warnings
import numpy as np
import pandas as pd
from scipy.stats import norm,chi2
import statsmodels.api as sm
from statsmodels.base.model import LikelihoodModel,LikelihoodModelResults,ResultMixin
from statsmodels.tools.decorators import cache_readonly
from statsmodels.tools.numdiff import approx_fprime

#%% GenericLikelihoodModel_TobitTruncreg

class GenericLikelihoodModel_TobitTruncreg(LikelihoodModel):
    """
    This is based on GenericLikelihoodModel of statsmodels (0.9.0).
    The modified part is indecated: see the following
        * `cov_type`
        * `genericmlefit`
    The aim of this class: to make it possible to use GenericLikelihoodModelResults_TobitTruncreg (see below)
    """
    def __init__(self, endog, exog=None, loglike=None, score=None,
                 hessian=None, missing='none', extra_params_names=None,
                 **kwds):

        if loglike is not None:
            self.loglike = loglike
        if score is not None:
            self.score = score
        if hessian is not None:
            self.hessian = hessian

        self.__dict__.update(kwds)

        super(GenericLikelihoodModel_TobitTruncreg, self).__init__(endog, exog,
                                                     missing=missing)

        if exog is not None:
            self.nparams = (exog.shape[1] if np.ndim(exog) == 2 else 1)

        if extra_params_names is not None:
            self._set_extra_params_names(extra_params_names)

    def _set_extra_params_names(self, extra_params_names):
        # check param_names
        if extra_params_names is not None:
            if self.exog is not None:
                self.exog_names.extend(extra_params_names)
            else:
                self.data.xnames = extra_params_names

        self.nparams = len(self.exog_names)

    def initialize(self):
        """
        Initialize (possibly re-initialize) a Model instance. For
        instance, the design matrix of a linear model may change
        and some things must be recomputed.
        """
        if not self.score:
            self.score = lambda x: approx_fprime(x, self.loglike)
            if not self.hessian:
                pass
        else:
            if not self.hessian:
                pass

        if self.exog is not None:
            er = np.linalg.matrix_rank(self.exog)
            self.df_model = float(er - 1)
            self.df_resid = float(self.exog.shape[0] - er)
        else:
            self.df_model = np.nan
            self.df_resid = np.nan
        super(GenericLikelihoodModel_TobitTruncreg, self).initialize()

    def expandparams(self, params):
        """
        expand to full parameter array when some parameters are fixed

        Parameters
        ----------
        params : array
            reduced parameter array

        Returns
        -------
        paramsfull : array
            expanded parameter array where fixed parameters are included

        Notes
        -----
        Calling this requires that self.fixed_params and self.fixed_paramsmask
        are defined.

        *developer notes:*

        This can be used in the log-likelihood to ...

        this could also be replaced by a more general parameter
        transformation.

        """
        paramsfull = self.fixed_params.copy()
        paramsfull[self.fixed_paramsmask] = params
        return paramsfull

    def reduceparams(self, params):
        """Reduce parameters"""
        return params[self.fixed_paramsmask]

    def loglike(self, params):
        """Log-likelihood of model at params"""
        return self.loglikeobs(params).sum(0)

    def nloglike(self, params):
        """Negative log-likelihood of model at params"""
        return -self.loglikeobs(params).sum(0)

    def loglikeobs(self, params):
        """Log-likelihood of individual observations at params"""
        return -self.nloglikeobs(params)

    def score(self, params):
        """
        Gradient of log-likelihood evaluated at params
        """
        kwds = {}
        kwds.setdefault('centered', True)
        return approx_fprime(params, self.loglike, **kwds).ravel()

    def score_obs(self, params, **kwds):
        """
        Jacobian/Gradient of log-likelihood evaluated at params for each
        observation.
        """
        kwds.setdefault('centered', True)
        return approx_fprime(params, self.loglikeobs, **kwds)

    def hessian(self, params):
        """
        Hessian of log-likelihood evaluated at params
        """
        from statsmodels.tools.numdiff import approx_hess
        # need options for hess (epsilon)
        return approx_hess(params, self.loglike)

    def fit(self, cov_type=None, start_params=None, method='nm', maxiter=500, full_output=1,   # modified
            disp=1, callback=None, retall=0, **kwargs):
        """
        Fit the model using maximum likelihood.

        The rest of the docstring is from
        statsmodels.LikelihoodModel.fit
        """
        if cov_type is None:          # modified
            cov_type = 'nonrobust'    # modified

        if start_params is None:
            if hasattr(self, 'start_params'):
                start_params = self.start_params
            else:
                start_params = 0.1 * np.ones(self.nparams)

        fit_method = super(GenericLikelihoodModel_TobitTruncreg, self).fit
        mlefit = fit_method(cov_type=cov_type, start_params=start_params,   # modified
                            method=method, maxiter=maxiter,
                            full_output=full_output,
                            disp=disp, callback=callback, **kwargs)

        # addition for modifincation starts ------------------------------------
        # if 'Truncreg' in str(self.model.__class__):
        if 'Truncreg' in str(self.__class__):
            genericmlefit = GenericLikelihoodModelResults_Truncreg(self, mlefit, cov_type)
        elif 'Tobit' in str(self.__class__):
            genericmlefit = GenericLikelihoodModelResults_Tobit(self, mlefit, cov_type)
        else:
            pass
        # addition for modifincation ends --------------------------------------

        exog_names = [] if (self.exog_names is None) else self.exog_names
        k_miss = len(exog_names) - len(mlefit.params)
        if not k_miss == 0:
            if k_miss < 0:
                self._set_extra_params_names(['par%d' % i
                                              for i in range(-k_miss)])
            else:
                import warnings
                warnings.warn('more exog_names than parameters', ValueWarning)

        return genericmlefit
#-----

#%% CommonAttributes_TobitTruncreg

class CommonAttributes_TobitTruncreg(object):
    # # Wald Test - all slopes=0
    @cache_readonly
    def wald_test_all_slopes(self):
        _mat = np.eye(len(self.params))[1:-1,:]
        return self.wald_test(_mat)

    @cache_readonly
    def prsquared(self):
        """
        McFadden's pseudo-R-squared. `1 - (llf / llnull)`
        """
        return 1 - self.llf/self.llnull

    @cache_readonly
    def llr(self):
        """
        Likelihood ratio chi-squared statistic; `-2*(llnull - llf)`
        under H0: all coefficients excluding constnat is zero
        """
        return -2*(self.llnull - self.llf)

    @cache_readonly
    def llr_pvalue(self):
        """
        p-value of likelihood ratio chi-squared statistic; `-2*(llnull - llf)`
        with degrees of freedom `df_model`
        under H0: all coefficients excluding constnat is zero
        """
        return chi2.sf(self.llr, self.df_model)

    @cache_readonly
    def llnull(self):
        """
        Value of loglikelihood with an Intercept only (no slope coefficients)
        """
        return self.result_null.llf

    def set_null_options(self, llnull=None, attach_results=True, **kwds):
        """set fit options for Null (constant-only) model

        This resets the cache for related attributes which is potentially
        fragile. This only sets the option, the null model is estimated
        when llnull is accessed, if llnull is not yet in cache.

        Parameters
        ----------
        llnull : None or float
            If llnull is not None, then the value will be directly assigned to
            the cached attribute "llnull".
        attach_results : bool
            Sets an internal flag whether the results instance of the null
            model should be attached. By default without calling this method,
            thenull model results are not attached and only the loglikelihood
            value llnull is stored.
        kwds : keyword arguments
            `kwds` are directly used as fit keyword arguments for the null
            model, overriding any provided defaults.

        Returns
        -------
        no returns, modifies attributes of this instance

        """
        self._cache.pop('llnull', None)
        self._cache.pop('llr', None)
        self._cache.pop('llr_pvalue', None)
        self._cache.pop('prsquared', None)
        if hasattr(self, 'result_null'):  # Tetsu
            del self.result_null          # Tetsu

        if llnull is not None:
            self._cache['llnull'] = llnull
        self._attach_nullmodel = attach_results
        self._optim_kwds_null = kwds

    @cache_readonly
    def result_null(self):
        """
        Value of loglikelihood with an Intercept only (no slope coefficients)
        """
        model = self.model
        kwds = model._get_init_kwds().copy()
        for key in getattr(model, '_null_drop_keys', []):
            del kwds[key]

        # the following lines are modified
        if 'Truncreg' in str(model.__class__):
            mod_null = model.__class__(model.endog, np.ones(self.nobs), left=model.left, right=model.right, **kwds)
        elif 'Tobit' in str(model.__class__):
            mod_null = model.__class__(model.endog, np.ones(self.nobs), cens=model.cens, left=model.left, right=model.right, **kwds)
        else:
            pass

        optim_kwds = getattr(self, '_optim_kwds_null', {}).copy()

        if 'start_params' in optim_kwds:

            sp_null = optim_kwds.pop('start_params')
        elif hasattr(model, '_get_start_params_null'):

            sp_null = model._get_start_params_null()
        else:
            sp_null = None

        opt_kwds = dict(method='bfgs', warn_convergence=False, maxiter=10000,disp=0)
        opt_kwds.update(optim_kwds)

        if optim_kwds:
            res_null = mod_null.fit(start_params=sp_null, **opt_kwds)
        else:
            res_null = mod_null.fit(start_params=sp_null, method='nm',
                                    warn_convergence=False,
                                    maxiter=10000, disp=0)

        if getattr(self, '_attach_nullmodel', False) is not False:
            self.res_null = res_null

        return res_null
#-----

#%% AdditionalAttributes_Truncreg

class AdditionalAttributes_Truncreg(object):
    @cache_readonly
    def fittedvalues(self):
        """
        y_hat (Linear fitted values)
        """
        return self.exog @ self.params[:-1].T

    # residuals
    @cache_readonly
    def resid(self):
        """
        Residuals = y - y_hat
        """
        return self.endog - self.fittedvalues


    @cache_readonly
    def fitted_endog(self):
        """
        E(y|x, cond)
            cond for left-truncated: y > left-truncated value
            cond for left-truncated: y < right-truncated value
            cond for left- & right-truncated: left-truncated value < y < right-truncated value
        Non-linear fitted endog variables (conditional expectations)
        But, this attribute may be that useful
        """
        s = self.params[-1]
        sigma = np.exp(s)
        Xb = self.fittedvalues
        _l = self.model.left
        _r = self.model.right

        if ~np.isneginf(_l) & np.isposinf(_r):
            first_term = Xb*norm.cdf(Xb, loc=_l, scale=sigma)
            second_term = sigma*norm.pdf(Xb, loc=_l, scale=sigma)
            return first_term + second_term

        elif np.isneginf(_l) & ~np.isposinf(_r):
            first_term = Xb*norm.sf(Xb, loc=_r, scale=sigma)
            second_term = sigma*norm.pdf(Xb, loc=_r, scale=sigma)
            return first_term - second_term

        elif ~np.isneginf(_l) & ~np.isposinf(_r):
            first_term = Xb*norm.cdf(Xb, loc=_l, scale=sigma)
            second_term = Xb*norm.cdf(Xb, loc=_r, scale=sigma)
            third_term = sigma*( norm.pdf(Xb, loc=_l, scale=sigma) - norm.pdf(Xb, loc=_r, scale=sigma) )
            return first_term - second_term + third_term

        else:
            warnings.warn('\n\n**********************************************************************\n\n'+
                          'Equivalent to untruncated Maximum Likelihood Estimation\n\n'+
                          '**********************************************************************\n')

#-----

#%% AdditionalAttributes_Tobit

class AdditionalAttributes_Tobit(object):
    @cache_readonly
    def fittedvalues(self):
        """
        y_hat (linear fiited latent variable)
        """
        return self.exog @ self.params[:-1].T

    # residuals
    @cache_readonly
    def resid(self):
        """
        Residuals = y - y_hat
        """
        return self.endog - self.fittedvalues

    # non-linear fitted endog variables (conditional expectations)
    @cache_readonly
    def fitted_endog(self):
        """
        E(y|x, cond)
            cond for left-truncated: y > left-truncated value
            cond for left-truncated: y < right-truncated value
            cond for left- & right-truncated: left-truncated value < y < right-truncated value
        Non-linear fitted endog variables (conditional expectations)
        """
        s = self.params[-1]
        sigma = np.exp(s)
        Xb = self.fittedvalues
        _l = self.model.left
        _r = self.model.right
        check_left = self.model.cens.unique().min()
        check_right = self.model.cens.unique().max()

        if (check_left == -1) & (check_right == 0):
            first_term = (Xb-_l)*norm.cdf(Xb, loc=_l, scale=sigma)
            second_term = sigma*norm.pdf(Xb, loc=_l, scale=sigma)
            return _l + first_term + second_term

        elif (check_left == 0) & (check_right == 1):
            first_term = (Xb-_r)*norm.sf(Xb, loc=_r, scale=sigma)
            second_term = sigma*norm.pdf(Xb, loc=_r, scale=sigma)
            return _r + first_term - second_term

        elif (check_left == -1) & (check_right == 1):
            first_term = (Xb-_l)*norm.cdf(Xb, loc=_l, scale=sigma)
            second_term = (Xb-_r)*norm.cdf(Xb, loc=_r, scale=sigma)
            third_term = sigma*( norm.pdf(Xb, loc=_l, scale=sigma) - norm.pdf(Xb, loc=_r, scale=sigma) )
            return _l + first_term - second_term + third_term

        else:
            warnings.warn('\n\n**********************************************************************\n\n'+
                          'Equivalent to fitted_endog of uncensored Maximum Likelihood Estimation\n\n'+
                          '**********************************************************************\n')


    # censoring
    @cache_readonly
    def obs(self):
        """
        (no of obs, no of uncensored obs, no of left-censored obs, no of right-censored obs)
        """
        censor = self.model.cens
        num_minus_1 = len(censor[censor == -1])
        num_0 = len(censor[censor == 0])
        num_1 = len(censor[censor == 1])
        uncensored = num_0
        left_censored = num_minus_1
        right_censored = num_1
        return self.nobs,uncensored,left_censored,right_censored


#-----

#%% GenericLikelihoodModelResults_Truncreg

class GenericLikelihoodModelResults_Truncreg(LikelihoodModelResults,
                                             ResultMixin,
                                             CommonAttributes_TobitTruncreg,
                                             AdditionalAttributes_Truncreg):
    """
    This is based on GenericLikelihoodModelResults of statsmodels (0.9.0).
    Some attributes are added.

    Attributes
    ----------
    aic : float
        Akaike information criterion.  -2*(`llf` - p) where p is the number
        of regressors including the intercept.
    bic : float
        Bayesian information criterion. -2*`llf` + ln(`nobs`)*p where p is the
        number of regressors including the intercept.
    bse : array
        The standard errors of the coefficients.
    df_resid : float
        See model definition.
    df_model : float
        See model definition.
    fitted_values : array
        Linear predictor XB.
    llf : float
        Value of the loglikelihood
    llnull : float
        Value of the constant-only loglikelihood
    llr : float
        Likelihood ratio chi-squared statistic; -2*(`llnull` - `llf`)
    llr_pvalue : float
        The chi-squared probability of getting a log-likelihood ratio
        statistic greater than llr.  llr has a chi-squared distribution
        with degrees of freedom `df_model`.
    prsquared : float
        McFadden's pseudo-R-squared. 1 - (`llf`/`llnull`)
    result_null:
        result under H0: all slope coefficients are zero
    """

    def __init__(self, model, mlefit, cov_type):
        self.model = model
        self.endog = model.endog
        self.exog = model.exog
        self.nobs = model.endog.shape[0]
        self.cov_type = cov_type         # added

        if hasattr(model, 'df_model'):
            self.df_model = model.df_model
        else:
            self.df_model = len(mlefit.params)

            self.model.df_model = self.df_model

        if hasattr(model, 'df_resid'):
            self.df_resid = model.df_resid
        else:
            self.df_resid = self.endog.shape[0] - self.df_model

            self.model.df_resid = self.df_resid

        self._cache = {}
        self.__dict__.update(mlefit.__dict__)


    def summary(self, yname=None, xname=None, title=None, alpha=.05):
        """Summarize the Regression Results

        Parameters
        ----------
        yname : string, optional
            Default is `y`
        xname : list of strings, optional
            Default is `var_##` for ## in p the number of regressors
        title : string, optional
            Title for the top table. If not None, then this replaces the
            default title
        alpha : float
            significance level for the confidence intervals

        Returns
        -------
        smry : Summary instance
            this holds the summary tables and text, which can be printed or
            converted to various output formats.

        See Also
        --------
        statsmodels.iolib.summary.Summary : class to hold summary results
        """

        # if 'truncreg' in str(self.model.__module__):
        top_left = [('Dep. Variable:', None),
                    ('Model:', None),
                    ('Method:', ['Maximum Likelihood']),
                    ('Date:', None),
                    ('Time:', None),
                    ('No. Observations:', None),
                    ('Df Residuals:', None),
                    ('Df Model:', None),
                    ]

        top_right = [('Pseudo R-squ:',["{:.3f}".format(self.prsquared)]),
                     ('Log-Likelihood:', ["{:.1f}".format(self.llf)]),
                     ('LL-Null:', ["{:.1f}".format(self.llnull)]),
                     ('LL-Ratio:', ["{:.1f}".format(self.llr)]),
                     ('LLR p-value:', ["{:.3f}".format(self.llr_pvalue)]),
                     ('AIC:', ["{:.1f}".format(self.aic)]),
                     ('BIC:', ["{:.1f}".format(self.bic)]),
                     ('Covariance Type:', [self.cov_type]),
                     ]

        if title is None:
            title = self.model.__class__.__name__ + ' ' + "Regression Results"

        # create summary table instance
        from statsmodels.iolib.summary import Summary
        smry = Summary()
        smry.add_table_2cols(self, gleft=top_left, gright=top_right,
                             yname=yname, xname=xname, title=title)
        smry.add_table_params(self, yname=yname, xname=xname, alpha=alpha,
                              use_t=self.use_t)

        return smry
#-----

#%% GenericLikelihoodModelResults_Tobit

class GenericLikelihoodModelResults_Tobit(LikelihoodModelResults,
                                          ResultMixin,
                                          CommonAttributes_TobitTruncreg,
                                          AdditionalAttributes_Tobit):
    """
    This is based on GenericLikelihoodModelResults of statsmodels (0.9.0).
    Some attributes are added.

    Attributes
    ----------
    aic : float
        Akaike information criterion.  -2*(`llf` - p) where p is the number
        of regressors including the intercept.
    bic : float
        Bayesian information criterion. -2*`llf` + ln(`nobs`)*p where p is the
        number of regressors including the intercept.
    bse : array
        The standard errors of the coefficients.
    df_resid : float
        See model definition.
    df_model : float
        See model definition.
    fitted_values : array
        Linear predictor XB.
    llf : float
        Value of the loglikelihood
    llnull : float
        Value of the constant-only loglikelihood
    llr : float
        Likelihood ratio chi-squared statistic; -2*(`llnull` - `llf`)
    llr_pvalue : float
        The chi-squared probability of getting a log-likelihood ratio
        statistic greater than llr.  llr has a chi-squared distribution
        with degrees of freedom `df_model`.
    prsquared : float
        McFadden's pseudo-R-squared. 1 - (`llf`/`llnull`)
    result_null:
        result under H0: all slope coefficients are zero
    """

    def __init__(self, model, mlefit, cov_type):
        self.model = model
        self.endog = model.endog
        self.exog = model.exog
        self.nobs = model.endog.shape[0]
        self.cov_type = cov_type         # added

        if hasattr(model, 'df_model'):
            self.df_model = model.df_model
        else:
            self.df_model = len(mlefit.params)

            self.model.df_model = self.df_model

        if hasattr(model, 'df_resid'):
            self.df_resid = model.df_resid
        else:
            self.df_resid = self.endog.shape[0] - self.df_model

            self.model.df_resid = self.df_resid

        self._cache = {}
        self.__dict__.update(mlefit.__dict__)


    def summary(self, yname=None, xname=None, title=None, alpha=.05):
        """Summarize the Regression Results

        Parameters
        ----------
        yname : string, optional
            Default is `y`
        xname : list of strings, optional
            Default is `var_##` for ## in p the number of regressors
        title : string, optional
            Title for the top table. If not None, then this replaces the
            default title
        alpha : float
            significance level for the confidence intervals

        Returns
        -------
        smry : Summary instance
            this holds the summary tables and text, which can be printed or
            converted to various output formats.

        See Also
        --------
        statsmodels.iolib.summary.Summary : class to hold summary results
        """

        top_left = [('Dep. Variable:', None),
                    # ('Model:', None),
                    ('Method:', ['Maximum Likelihood']),
                    # ('Date:', None),
                    # ('Time:', None),
                    ('No. Observations:', None),
                    ('No. Uncensored Obs:', ["{:.0f}".format(self.obs[1])]),
                    ('No. Left-censored Obs:', ["{:.0f}".format(self.obs[2])]),
                    ('No. Right-censored Obs:', ["{:.0f}".format(self.obs[3])]),
                    ('Df Residuals:', None),
                    ('Df Model:', None),
                    ]

        top_right = [('Pseudo R-squ:',["{:.3f}".format(self.prsquared)]),
                     ('Log-Likelihood:', ["{:.1f}".format(self.llf)]),
                     ('LL-Null:', ["{:.1f}".format(self.llnull)]),
                     ('LL-Ratio:', ["{:.1f}".format(self.llr)]),
                     ('LLR p-value:', ["{:.3f}".format(self.llr_pvalue)]),
                     ('AIC:', ["{:.1f}".format(self.aic)]),
                     ('BIC:', ["{:.1f}".format(self.bic)]),
                     ('Covariance Type:', [self.cov_type]),
                     ]

        if title is None:
            title = self.model.__class__.__name__ + ' ' + "Regression Results"

        # create summary table instance
        from statsmodels.iolib.summary import Summary
        smry = Summary()
        smry.add_table_2cols(self, gleft=top_left, gright=top_right,
                             yname=yname, xname=xname, title=title)
        smry.add_table_params(self, yname=yname, xname=xname, alpha=alpha,
                              use_t=self.use_t)

        return smry
#-----

# EOF
